/************************************************************************/
/*									*/
/*			Connect several AVRs over one wire              */
/*		Author: Peter Dannegger					*/
/*									*/
/************************************************************************/
//                      Target: ATmega48
/*
Master:         ATmega644
Address:        none
Data pin:       PB3 = OC0A (transmit) = PCINT11 (receive)
KEY0:           PB4
KEY1:           PB5
KEY2:           PB6
LED0:           PC1
LED1:           PC2
LED2:           PC3

Slave 1:        ATmega48
Address:        1
Data pin:       PB3 = OC2A (transmit) = PCINT3 (receive)
KEY0:           PB0
KEY1:           PB1
LED0:           PC0

Slave 2:        ATtiny261
Address:        2
Data pin:       PB3 = OC1B (transmit) = PCINT11 (receive)
KEY0:           PB2
LED0:           PA6
LED1:           PA7

Data Protocol:
1 * T: o-Bit
2 * T: 1-Bit
4 * T: Sync-Bit
*/

#include <avr/interrupt.h>
#include <avr/power.h>
#include "mydefs.h"

#define F_CPU           8e6                     // 8MHz
#define BIT_CLK         10e3                    // Bit time T = 10kHz
#define PRESCALER       64                      // F_CPU / 64
#define PRESCALER_SET	(1<<CS22)
#define	BIT_TIME	(F_CPU / BIT_CLK / PRESCALER)

#define DATA_SIZE       2                       // 2 Byte

#define	KEY0_PIN        SBIT( PINB,  PB0 )
#define	KEY1_PIN        SBIT( PINB,  PB1 )
#define	LED0            SBIT( PORTC, PC0 )
#define	LED0_DDR        SBIT( DDRC,  PC0 )
#define	DATA            SBIT( PORTB, PB3 )
#define	DATA_DDR        SBIT( DDRB,  PB3 )
#define	DATA_PIN        SBIT( PINB,  PB3 )


/****************************** Transmission ****************************/

uint8_t tx_buff[DATA_SIZE];
uint8_t tx_idx;


ISR( TIMER2_COMPA_vect )                        // Transmit interrupt
{
  uint8_t *dp;					// locals for faster access

  if( tx_idx < DATA_SIZE * 8 ){
    OCR2A = BIT_TIME * 1 - 0.5;                 // 0-bit: 1T
    dp = &tx_buff[tx_idx++ / 8];                // get byte address
    if( *dp & 0x80 )                            // msb first
      OCR2A = BIT_TIME * 2 - 0.5;               // 1-bit: 2T
    *dp <<= 1;                                  // shift bits
  }else{
    OCR2A = BIT_TIME * 4 - 0.5;                 // stop:  4T
    if( DATA_PIN )                              // after output = high
      TCCR2B = 0;                               // then stop T1
  }
}


void tx_data( uint8_t addr, uint8_t data )
{
  PCMSK0 = 0;    				// disable receive
  TCCR2B = 0;					// stop T1
  volatil(tx_idx) = 0;                          // point to first bit
  volatil(tx_buff[0]) = addr;
  volatil(tx_buff[1]) = data;
  DATA_DDR = 0;                                 // input
  DATA = 1;                                     // pullup on
  while( DATA_PIN == 0 );                       // wait until pin high
  TCNT2 = 0;
  OCR2A = BIT_TIME * 4 - 0.5;                   // sync: 4T
  TCCR2A = 1<<COM2A1 | 1<<COM2A0;               // set on compare
  TCCR2B = 1<<FOC2A;                            // set output high
  TCCR2A = 1<<COM2A0                            // toggle pin on compare
         | 1<<WGM21;                            // Mode: CTC
  DATA_DDR = 1;                                 // tx output on
  TIFR2 = 1<<OCF2A;                             // clear earlier interrupts
  TIMSK2 = 1<<OCIE2A;                           // enable tx interrupt
  TCCR2B = PRESCALER_SET;                       // start T2
  while( TCCR2B );                              // until tx done
}


/****************************** Receiving *******************************/

#define rx_idx  tx_idx
#define rx_buff tx_buff


ISR( PCINT0_vect )                              // Receive interrupt
{
  uint8_t *dp;					// locals for faster access
  uint8_t i;                                	// locals for faster access

  TCNT2 = 0;
  if( rx_idx >= DATA_SIZE * 8 ){                // on start or stop bit
    rx_idx = 0;                                 // point to first bit
  }else{
    dp = &rx_buff[rx_idx / 8];                  // get byte address
    i = *dp << 1;                               // shift bits
    if( TIFR2 & 1<<OCF2A )                      // if pulse > 1.4T
      i++;                                      // msb first
    *dp = i;
    rx_idx++;                                   // point to next bit
  }
  TIFR2 = 1<<OCF2A;                             // clear 2T detection flag
}


ISR( TIMER2_COMPB_vect )                        // sync received
{
  if( rx_idx == DATA_SIZE * 8 )                 // receive successful
    PCMSK0 = 0;					// disable further receive
  rx_idx = 255;                                 // detect sync transition
}


#define	rx_busy() PCMSK0                        // 0 = data received


void rx_init( void )
{
  TCCR2B = 0;					// Stop T0
  volatil(rx_idx) = 255;                        // detect sync transition
  DATA_DDR = 0;                                 // input
  DATA = 1;                                     // pullup on
  TCCR2A = 0;					// disconnect tx output
  TCNT2 = 0;
  OCR2A = BIT_TIME * 1.4 - 0.5;                 // detect 2T
  OCR2B = BIT_TIME * 2.8 - 0.5;                 // detect 4T
  TIFR2 = 1<<OCF2A | 1<<OCF2B;                  // clear earlier interrupts
  TIMSK2 = 1<<OCIE2B;                           // start/stop bit interrupt
  TCCR2B = PRESCALER_SET;                       // start T2
  PCMSK0 = 1<<PCINT3;
  PCIFR = 1<<PCIF0;                             // clear earlier interrupts
  PCICR = 1<<PCIE0;                             // pin change interrupt
}


int main( void )
{
  clock_prescale_set( clock_div_1 );
  PORTB = 0xFF;					// input with pullup
  PORTC = 0xFF;					// input with pullup
  PORTD = 0xFF;					// input with pullup
  LED0 = 0;
  LED0_DDR = 1;                                 // Output
  
  sei();
  for(;;){
    rx_init();                                  // start receive
    while( rx_busy() );
    if( volatil(rx_buff[0]) == 1 ){             // if slave 1 addressed
      LED0 = volatil(rx_buff[1]) & 1;
      tx_data( 1, KEY0_PIN | KEY1_PIN<<1 );     // answer from slave 1
    }
  }
}
